/* -*- C++ -*- vim: set syntax=cpp: 
 * PURPOSE: Token Queue 
 *
 * A token queue is a queue where read and write cycles are separate.
 * That is, when the queue is filled, it is not read until the 
 * filling terminated. Then, the read does not terminate before there
 * is no more token left.
 *
 * Wrap-arround is neither necessary nor meaningful!
 *
 * (C) 2004-2017 Frank-Rene Schaefer                                          */
#ifndef QUEX_INCLUDE_GUARD_TestAnalyzer__TOKEN__TOKEN_QUEUE_I
#define QUEX_INCLUDE_GUARD_TestAnalyzer__TOKEN__TOKEN_QUEUE_I

#include "test_cpp_cg/lib/definitions"
#include "test_cpp_cg/lib/quex/asserts"
#include "test_cpp_cg/lib/quex/MemoryManager"

/* NOTE: TestAnalyzer_Token must be defined at this place!                       */

#include "test_cpp_cg/lib/token/TokenQueue"




#define QUEX_TOKEN_ID_IS_REPEATABLE(ID) ((ID) == QUEX_TKN_X || (ID) == QUEX_TKN_DEDENT) 



inline void             
TestAnalyzer_TokenQueue_push_core(TestAnalyzer_TokenQueue* me, TestAnalyzer_token_id_t Id);

inline bool
TestAnalyzer_TokenQueue_construct(TestAnalyzer_TokenQueue* me, 
                                TestAnalyzer*    lexer,
                                const size_t           N)
/* me:     The token queue.
 * Memory: Pointer to memory of token queue, 0x0 --> no initial memory.
 * N:      Number of token objects that the array can carry.                  */
{
    TestAnalyzer_Token*   iterator   = 0x0;
    TestAnalyzer_Token*   memory     = (TestAnalyzer_Token*)quex::MemoryManager_allocate(
                                             N * sizeof(TestAnalyzer_Token),
                                             E_MemoryObjectType_TOKEN_ARRAY);
    TestAnalyzer_Token*   memory_end = &memory[N];

    if( ! memory ) {
        TestAnalyzer_TokenQueue_resources_absent_mark(me);
        return false;
    }

    __quex_assert(memory != 0x0);

    /* Call placement new (plain constructor) for all tokens in chunk.        */
    for(iterator = memory; iterator != memory_end; ++iterator) {
        TestAnalyzer_Token_construct(iterator);
    }
    TestAnalyzer_TokenQueue_init(me, memory, memory_end); 
    me->the_lexer = lexer;
    return true;
}

inline void
TestAnalyzer_TokenQueue_reset(TestAnalyzer_TokenQueue* me) 
{                                                    
    me->read_iterator  = (TestAnalyzer_Token*)me->begin; 
    me->write_iterator = (TestAnalyzer_Token*)me->begin; 
}

inline void
TestAnalyzer_TokenQueue_init(TestAnalyzer_TokenQueue* me, 
                           TestAnalyzer_Token*       Memory, 
                           TestAnalyzer_Token*       MemoryEnd) 
{
    me->begin = Memory;                           
    me->end   = MemoryEnd;                        
    TestAnalyzer_TokenQueue_reset(me);                                
}

inline void
TestAnalyzer_TokenQueue_resources_absent_mark(TestAnalyzer_TokenQueue* me) 
{
    me->begin          = (TestAnalyzer_Token*)0;                           
    me->end            = (TestAnalyzer_Token*)0;
    me->read_iterator  = (TestAnalyzer_Token*)0; 
    me->write_iterator = (TestAnalyzer_Token*)0; 
    me->the_lexer      = (TestAnalyzer*)0; 
}

inline bool
TestAnalyzer_TokenQueue_resources_absent(TestAnalyzer_TokenQueue* me) 
{
    return    me->begin          == (TestAnalyzer_Token*)0                           
           && me->end            == (TestAnalyzer_Token*)0
           && me->read_iterator  == (TestAnalyzer_Token*)0 
           && me->write_iterator == (TestAnalyzer_Token*)0
           && me->the_lexer      == (TestAnalyzer*)0;
}

inline void
TestAnalyzer_TokenQueue_destruct(TestAnalyzer_TokenQueue* me)
{
    TestAnalyzer_Token* iterator = 0x0;
    /* Call explicit destructors for all tokens in array                      */
    for(iterator = me->begin; iterator != me->end; ++iterator) {
        TestAnalyzer_Token_destruct(iterator);
    }

    quex::MemoryManager_free((void*)&me->begin[0],
                               E_MemoryObjectType_TOKEN_ARRAY);

    /* The memory chunk for the token queue itself is located inside the
     * analyzer object. Thus, no explicit free is necessary. In case of user
     * managed token queue memory the user takes care of the deletion.        */
    TestAnalyzer_TokenQueue_resources_absent_mark(me);
}

inline void   
TestAnalyzer_TokenQueue_remainder_get(TestAnalyzer_TokenQueue* me,
                                    TestAnalyzer_Token**      begin,
                                    TestAnalyzer_Token**      end)
{
    *begin = me->read_iterator;
    *end   = me->write_iterator;
    TestAnalyzer_TokenQueue_reset(me);
}

inline void 
TestAnalyzer_TokenQueue_memory_get(TestAnalyzer_TokenQueue* me,
                                 TestAnalyzer_Token**      memory,
                                 size_t*                n)
{
    *memory = me->begin;
    *n      = (size_t)(me->end - me->begin);
}

inline bool 
TestAnalyzer_TokenQueue_is_full(TestAnalyzer_TokenQueue* me) 
{ return me->write_iterator >= me->end; }

inline bool 
TestAnalyzer_TokenQueue_is_empty(TestAnalyzer_TokenQueue* me)
{ return me->read_iterator == me->write_iterator; }

inline void             
TestAnalyzer_TokenQueue_push(TestAnalyzer_TokenQueue* me,
                           TestAnalyzer_token_id_t     Id)
/* Push a token and set only its token identifier.                           */
{

    if( QUEX_TOKEN_ID_IS_REPEATABLE(Id) ) {
        me->write_iterator->number = 1;
    }

    TestAnalyzer_TokenQueue_push_core(me, Id);
}

inline void             
TestAnalyzer_TokenQueue_push_core(TestAnalyzer_TokenQueue* me,
                                TestAnalyzer_token_id_t     Id)
{
    if( TestAnalyzer_TokenQueue_is_full(me) ) {
        me->the_lexer->error_code = E_Error_Token_QueueOverflow;
        return;
    }
    TestAnalyzer_TokenQueue_assert_before_sending(me);  

       me->write_iterator->line_n   = me->the_lexer->counter._line_number_at_begin;
     me->write_iterator->column_n = me->the_lexer->counter._column_number_at_begin;

    me->write_iterator->id = Id;              
    ++(me->write_iterator);       
}


inline void             
TestAnalyzer_TokenQueue_push_text(TestAnalyzer_TokenQueue* me,
                                TestAnalyzer_token_id_t     Id,
                                TestAnalyzer_lexatom_t*     BeginP,
                                TestAnalyzer_lexatom_t*     EndP)
/* Push a token and set its 'text' member.                                    */
{
    TestAnalyzer_TokenQueue_assert_before_sending(me);
    TestAnalyzer_Token_take_text(me->write_iterator, BeginP, EndP);
    TestAnalyzer_TokenQueue_push(me, Id);
}





inline void             
TestAnalyzer_TokenQueue_push_repeated(TestAnalyzer_TokenQueue* me,
                                    TestAnalyzer_token_id_t     Id,
                                    size_t                 RepetitionN)
/* Push a repeated token by 'RepetitionN' times. This is only addmissible for
 * TokenId-s specified in the 'repeated_token' section of the '.qx' file.     */
{
    TestAnalyzer_TokenQueue_assert_before_sending(me);  
    __quex_assert(RepetitionN != 0);        
    __quex_assert(QUEX_TOKEN_ID_IS_REPEATABLE(Id));

    me->write_iterator->number = RepetitionN;
    TestAnalyzer_TokenQueue_push_core(me, Id);
}


inline TestAnalyzer_Token* 
TestAnalyzer_TokenQueue_pop(TestAnalyzer_TokenQueue* me)
{
    __quex_assert(TestAnalyzer_TokenQueue_begin(me) != 0x0);

    if( TestAnalyzer_TokenQueue_is_empty(me) ) {        
        return (TestAnalyzer_Token*)0;
    }
    else if( QUEX_TOKEN_ID_IS_REPEATABLE(me->read_iterator->id) ) {
        if( me->read_iterator->number > 1 ) { 
            me->read_iterator->number -= 1;
            if( me->read_iterator->number != 0 ) {
                return me->read_iterator;  
            }
        }
    }
    /* Tokens are in queue --> take next token from queue                    */ 
    return me->read_iterator++;
}
#undef QUEX_TOKEN_ID_IS_REPEATABLE




inline TestAnalyzer_Token* TestAnalyzer_TokenQueue_begin(TestAnalyzer_TokenQueue* me)
{ return me->begin; }

inline TestAnalyzer_Token* TestAnalyzer_TokenQueue_back(TestAnalyzer_TokenQueue* me)
{ return me->end - 1; }

inline TestAnalyzer_Token* TestAnalyzer_TokenQueue_last_token(TestAnalyzer_TokenQueue* me)
{ return me->write_iterator == me->begin ? (TestAnalyzer_Token*)0 : &me->write_iterator[-1]; }

inline size_t TestAnalyzer_TokenQueue_available_n(TestAnalyzer_TokenQueue* me) 
{ return (size_t)(me->end - me->write_iterator); }

inline void
TestAnalyzer_TokenQueue_set_token_TERMINATION(TestAnalyzer_TokenQueue* me) 
/* Reset entire token queue and set the token 'TERMINATION'. This should
 * only be called in case of a detected error.                                */
{
    TestAnalyzer_TokenQueue_reset(me);

    TestAnalyzer_TokenQueue_push_text(me, QUEX_TKN_TERMINATION, 
                                    (TestAnalyzer_lexatom_t*)0, (TestAnalyzer_lexatom_t*)0);


}

inline bool             
TestAnalyzer_TokenQueue_assert_before_sending(TestAnalyzer_TokenQueue* me)
{                                                                              
    if( ! TestAnalyzer_TokenQueue_assert_after_sending(me) ) {
        return false;
    } 
    /* End of token queue has not been reached.                       */          
    __quex_assert((me)->write_iterator != (me)->end);                             
    /* No token sending after 'TERMINATION'.                          */          
    __quex_assert(   (me)->write_iterator         == (me)->begin                  
                  || (me)->write_iterator[-1].id !=  QUEX_TKN_TERMINATION ); 
    return true;
} 

inline bool             
TestAnalyzer_TokenQueue_assert_after_sending(TestAnalyzer_TokenQueue* me)
{                                                                     
    __quex_assert((me)->begin != 0x0);                                   
    __quex_assert((me)->read_iterator  >= (me)->begin);                  
    __quex_assert((me)->write_iterator >= (me)->read_iterator);          
    /* If the following breaks, then the given queue size was to small*/ 
    if( (me)->write_iterator > (me)->end ) {                            
        QUEX_ERROR_EXIT("Error: Token queue overflow.");                 
        return false;
    }                                                                    
    return true;
}



#endif /* QUEX_INCLUDE_GUARD_TestAnalyzer__TOKEN__TOKEN_QUEUE_I */
